/*
    2d cleaner filter for VirtualDub -- blends a pixel with pixels 
	surrounding it as long as those pixels are simular to the source
	pixel

	Current version -- Parashar Krishnamachari AKA CPIasminc
*/

#define VDEXT_VIDEO_FILTER
#define VDEXT_MAIN

#include "filter.h"
#include "resource.h"
#include <stdio.h>
#include <math.h>
#include "ScriptInterpreter.h"
#include "ScriptError.h"
#include "ScriptValue.h"
#include "VBitmap.h"

#define throwaway
#define newcalc
#define calcn
#define sep

extern HINSTANCE g_hInst;

typedef struct MyFilterData {
	float fThreshold;    // Accuracy
	int fRadius;		// Initial Spacing
} MyFilterData;

bool IsPowerOfTwo ( unsigned x );
float ratio2(float a, float b);
float ratio4(float a, float b, float c, float d);
static int twodclean_run(const FilterActivation *fa, const FilterFunctions *ff);
BOOL CALLBACK twodclean_ConfigDlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam);
int twodclean_ConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd);
void twodclean_StringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str);
void twodcleanScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc);
bool twodcleanFssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen);

ScriptFunctionDef twodclean_func_defs[]={
    { (ScriptFunctionPtr)twodcleanScriptConfig, "Config", "0iii" },
    { NULL },
};

CScriptObject twodclean_obj={
    NULL, twodclean_func_defs
};

typedef unsigned char byte;

static float fThreshold;
static int pitch;

bool IsPowerOfTwo ( unsigned x )
{
    if ( x < 2 )
        return FALSE;

    if ( x & (x-1) )
        return FALSE;

    return TRUE;
}


float ratio2(float a, float b)
{
    if ((a==0) && (b==0))    return 1.0;
    if ((a==0) || (b==0))    return 0.0;

    if (a>b)    return b/a;
    else        return a/b;
}


float ratio4(float a, float b, float c, float d) 
{
    float q1 = ratio2(a,b);
    float q2 = ratio2(c,d);

    if (q1 < q2)    
		return q1/q2;
    else          
		return q2/q1;
}


static int twodclean_run(const FilterActivation *fa, const FilterFunctions *ff) {
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	unsigned long w,h;
	unsigned long *src = (unsigned long *)fa->src.data;
	unsigned long *dst = new unsigned long[sizeof(unsigned long) * fa->src.w * fa->src.h];
	unsigned int x, y;
	pitch = fa->src.pitch;
	fThreshold = mfd->fThreshold;
	w = fa->src.w;
	h = fa->src.h;

	int Spacing = mfd->fRadius;
	unsigned char r1, r2, r3, r4;
	unsigned char g1, g2, g3, g4;
	unsigned char b1, b2, b3, b4;
	unsigned int HalfSpace;

	VBitmap *tmp = new VBitmap;
	tmp->init(dst, (PixDim)fa->dst.w, (PixDim)fa->dst.h, fa->dst.depth);
	dst = (unsigned long *)tmp->data;

	// Copy the source image...
	tmp->BitBlt(0,0,&(fa->src),0,0,-1,-1);

	while (Spacing > 1)
	{
		HalfSpace = Spacing >> 1;
		float CurT = (float)pow(fThreshold, HalfSpace);

		// This is the loop for all the two-neighbour interpolants...
		// Obviously, it doesn't matter if we're not going to interpolate, but
		// that's something that can only be determined at the pixel level.
		
		for (y = HalfSpace; y < (h+HalfSpace); y += HalfSpace)
		{
			for (x = HalfSpace; x < (w+HalfSpace); x += HalfSpace)
			{
				if (x < w)
				{
					int x1 = x;
					int y1 = y - HalfSpace;

					int ofs = w * y1 + x1 + HalfSpace;
					unsigned long p = src[ofs];
					r1 = (byte)((p & 0xFF0000) >> 16);
					g1 = (byte)((p & 0x00FF00) >> 8);
					b1 = (byte)(p & 0x00FF);

					ofs = w * y1 + x1 - HalfSpace;
					p = src[ofs];
					r2 = (byte)((p & 0xFF0000) >> 16);
					g2 = (byte)((p & 0x00FF00) >> 8);
					b2 = (byte)(p & 0x00FF);

					ofs = w * y1 + x1;
					// If the ratios are below the threshold, then we just do linear interp.
					// Otherwise, keep the original values.
					if ( (ratio2(r1,r2) <= CurT) && (ratio2(g1,g2) <= CurT) && (ratio2(b1,b2) <= CurT) )
					{
						unsigned long a = (r1 << 16) | (g1 << 8) | b1;
						unsigned long b = (r2 << 16) | (g2 << 8) | b2;
						dst[ofs] = ((a & 0xFEFEFE) >> 1) + ((b & 0xFEFEFE) >> 1) | ((a | b) & 0x010101);
					}
					else
					{
						p = src[ofs];
						dst[ofs] = p;
					}

				} // end if (x < w)
				
				if (y < h)
				{
					int x1 = x - HalfSpace;
					int y1 = y;

					int ofs = w * (y1 + HalfSpace) + x1;
					unsigned long p = src[ofs];
					r1 = (byte)((p & 0xFF0000) >> 16);
					g1 = (byte)((p & 0x00FF00) >> 8);
					b1 = (byte)(p & 0x00FF);

					ofs = w * (y1 - HalfSpace) + x1;
					p = src[ofs];
					r2 = (byte)((p & 0xFF0000) >> 16);
					g2 = (byte)((p & 0x00FF00) >> 8);
					b2 = (byte)(p & 0x00FF);

					ofs = w * y1 + x1;
					if ( (ratio2(r1,r2) <= CurT) && (ratio2(g1,g2) <= CurT) && (ratio2(b1,b2) <= CurT) )
					{
						unsigned long a = (r1 << 16) | (g1 << 8) | b1;
						unsigned long b = (r2 << 16) | (g2 << 8) | b2;
						dst[ofs] = ((a & 0xFEFEFE) >> 1) + ((b & 0xFEFEFE) >> 1) | ((a | b) & 0x010101);
					}
					else
					{
						p = src[ofs];
						dst[ofs] = p;
					}

				} // end if (y < h)
			} // end for loop x
		} // end for loop y

		// Now you have a loop for the 4-neighbour interpolants.
		for (y = HalfSpace; y <= (h-HalfSpace); y += HalfSpace)
		{
			for (x = HalfSpace; x <= (w - HalfSpace); x += HalfSpace)
			{
				int ofs = w * (y-HalfSpace) + x;
				unsigned long p = dst[ofs];
				r1 = (byte)((p & 0xFF0000) >> 16);
				g1 = (byte)((p & 0x00FF00) >> 8);
				b1 = (byte)(p & 0x00FF);

				ofs = w * (y + HalfSpace) + x;
				p = dst[ofs];
				r2 = (byte)((p & 0xFF0000) >> 16);
				g2 = (byte)((p & 0x00FF00) >> 8);
				b2 = (byte)(p & 0x00FF);

				ofs = w * y + x - HalfSpace;
				p = dst[ofs];
				r3 = (byte)((p & 0xFF0000) >> 16);
				g3 = (byte)((p & 0x00FF00) >> 8);
				b3 = (byte)(p & 0x00FF);

				ofs = w * y + x + HalfSpace;
				p = dst[ofs];
				r4 = (byte)((p & 0xFF0000) >> 16);
				g4 = (byte)((p & 0x00FF00) >> 8);
				b4 = (byte)(p & 0x00FF);

				ofs = w * y + x;
				if ( (ratio4(r1,r2,r3,r4) <= CurT) && (ratio4(g1,g2,g3,g4) <= CurT) && (ratio4(b1,b2,b3,b4) <= CurT) )
				{
					r1 = (byte)((int)(r1 + r2 + r3 + r4) >> 2);
					g1 = (byte)((int)(g1 + g2 + g3 + g4) >> 2);
					b1 = (byte)((int)(b1 + b2 + b3 + b4) >> 2);
					dst[ofs] = (r1 << 16) | (g1 << 8) | (b1);
				}
				else
				{
					p = src[ofs];
					dst[ofs] = p;
				}

			} // end for loop x
		} // end for loop y

		Spacing = HalfSpace;
	} // end while loop

	fa->dst.BitBlt(0,0,tmp,0,0,-1,-1);
	delete[] dst;
	delete tmp;
	return 0;
}

BOOL CALLBACK twodclean_ConfigDlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam) {
         MyFilterData *mfd = (MyFilterData *)GetWindowLong(hdlg, DWL_USER);

         switch(msg) {
             case WM_INITDIALOG:
                 SetWindowLong(hdlg, DWL_USER, lParam);
                 mfd = (MyFilterData *)lParam;
				 SetDlgItemInt(hdlg, IDC_THRESHOLD, (unsigned int)(mfd->fThreshold * 1000), FALSE);				
				 SetDlgItemInt(hdlg, IDC_RADIUS, mfd->fRadius, FALSE);

                 return TRUE;

             case WM_COMMAND:
                 switch(LOWORD(wParam)) {
                 case IDOK:
					// just added to fix cancel bug
					long threshold;
					BOOL success;

					threshold = GetDlgItemInt(hdlg, IDC_RADIUS, &success, FALSE);
					if (!success || threshold < 2 || !IsPowerOfTwo(threshold)) {
						SetFocus((HWND)lParam);
						MessageBeep(MB_ICONQUESTION);
						return TRUE;
					}

					mfd->fRadius = threshold;

					threshold = GetDlgItemInt(hdlg, IDC_THRESHOLD, &success, FALSE);
					if (!success || threshold <= 0 || threshold >= 1000) {
						SetFocus((HWND)lParam);
						MessageBeep(MB_ICONQUESTION);
						return TRUE;
					}

					mfd->fThreshold = (float)threshold / 1000.0f;                     
					
					 EndDialog(hdlg, 0);
                     return TRUE;
                 case IDCANCEL:
                     EndDialog(hdlg, 1);
                     return FALSE;
				 case IDC_RADIUS:
				if (HIWORD(wParam) == EN_KILLFOCUS) {
				}
				return TRUE;

				 case IDC_THRESHOLD:
				if (HIWORD(wParam) == EN_KILLFOCUS) {
				} 
				return TRUE;

                 }
                 break;
         }

         return FALSE;
     }


int twodclean_ConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd) {
         return DialogBoxParam(fa->filter->module->hInstModule,
                 MAKEINTRESOURCE(IDD_2DCLEAN), hwnd,
                 twodclean_ConfigDlgProc, (LPARAM)fa->filter_data);
     }

void twodclean_StringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str) {
         
         MyFilterData *mfd = (MyFilterData *)fa->filter_data;
		 sprintf(str, " ,threshold of %1.4f, Initial spacing of %dx%d)", 
			 mfd->fThreshold,
			 mfd->fRadius, mfd->fRadius);
     }

void twodcleanScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc) {
    FilterActivation *fa = (FilterActivation *)lpVoid;
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	mfd->fThreshold	= (float)argv[0].asInt() / 1000.0f;
	mfd->fRadius	= argv[1].asInt();
}

bool twodcleanFssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

    _snprintf(buf, buflen, "Config(%d, %d)",
		(int)(mfd->fThreshold * 1000.0f),
		mfd->fRadius);

    return true;
}

static int twodclean_start(FilterActivation *fa, const FilterFunctions *ff) {
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	return 0;
}

int twodclean_init(FilterActivation *fa, const FilterFunctions *ff) {
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	INITIALIZE_VTBLS;

	// These are decent defaults
	mfd->fThreshold = 0.5;
	mfd->fRadius = 4;
	return 0;
}

FilterDefinition filterDef_2dclean={
	0,0,NULL,
	"2d Cleaner",
	"Replaces each pixel with the average of its neighbors, as long as the average falls within a specified threshold",
	"CPIasminc",NULL,
	sizeof(MyFilterData),
	twodclean_init,NULL,
	twodclean_run,
	NULL,
	twodclean_ConfigProc,
	twodclean_StringProc,
	NULL,
	NULL,
	&twodclean_obj,          // script_obj
    twodcleanFssProc,        // fssProc

};

extern "C" int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat);
     extern "C" void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff);

     static FilterDefinition *fd_tutorial;

     int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat) {
         if (!(fd_tutorial = ff->addFilter(fm, &filterDef_2dclean, sizeof(FilterDefinition))))
             return 1;

         vdfd_ver    = VIRTUALDUB_FILTERDEF_VERSION;
         vdfd_compat = VIRTUALDUB_FILTERDEF_COMPATIBLE;

         return 0;
     }

     void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff) {
         ff->removeFilter(fd_tutorial);
     }

